/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * ContentDatabase (tdb implementation). 
 * @author Christian Grothoff
 * @author Uli Luckas
 * @file applications/afs/database/contentdatabase_tdb.c
 **/

#include "config.h"
#include "applications/afs/database/contentdatabase_tdb.h"
#include "util/xmalloc.h"
#include "util/logging.h"
#include "util/storage.h"
#include "util/returnvalues.h"
#include "util/statuscalls.h"

#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <pwd.h>
#include <grp.h>
#include <dirent.h>
#include <errno.h>

#ifdef TDB_STORAGE

#define TDB_DEBUG 0

static TDB_CONTEXT * dbf;

/**
 * Open a tdb database (for content)
 * @param dir the directory where content is configured 
 * to be stored (e.g. data/content). A file called ${dir}.dbf is used instead
 **/
static TDB_CONTEXT * getDatabase(char *dir) {
  char * fin;
  char * ff;
  TDB_CONTEXT *dbf;

  char *ext = ".tdb";
  int fnSize = strlen(dir) + strlen(ext);

#if TDB_DEBUG  
  LOG(LOG_DEBUG,
      "DEBUG: Database: %s (TDB)\n", 
      dir);
#endif
  ff = xmalloc(fnSize,
	       "getDatabase: filename");  
  strcpy(ff, dir);
  ff[strlen(ff) - 1] = 0;
  strcat(ff, ext);
  fin = expandFileName(ff);
  xfree(ff,
	"getDatabase: filename");
  
  dbf = tdb_open(fin, 0, 0, O_RDWR | O_CREAT, S_IRUSR|S_IWUSR);
  if (NULL == dbf) 
    errexit("getDatabase: failed to open database file %s with error: %s\n",
	    fin,
	    tdb_errorstr(dbf));
  registerSpaceMonitor(fin);
  xfree(fin,
	"getDatabase: filename (2)");
  
  return dbf;
}

void initContentDatabase_TDB(char * dir) {  
  dbf = getDatabase(dir);
}

/**
 * Clean shutdown of the storage module (not used at the moment)
 **/
void doneContentDatabase_TDB() {
  tdb_close(dbf);
}


/**
 * Call a method for each key in the database and
 * call the callback method on it. 
 *
 * @param callback the callback method
 * @param data second argument to all callback calls
 * @return the number of items stored in the content database
 **/
int forEachEntryInDatabase_TDB(EntryCallback callback,
			       void * data) {
  TDB_DATA prevkey, key;
  HashCode160 doubleHash;
  int count;

  count = 0;
  /* scan database data/content.dbf and add entries to database 
     if not already present */
  key = tdb_firstkey(dbf);
  while (key.dptr) {
    if (strlen(key.dptr) == sizeof(HashCode160)*2) {	
      if (callback != NULL) {
	hex2hash((HexName*)key.dptr,
		 &doubleHash);
	callback(&doubleHash, 
		 data);
      }
      count++; /* one more file */
    }
    prevkey = key;
    key = tdb_nextkey(dbf, prevkey);
    xfree(prevkey.dptr, "database dptr (allocation in DB code)");
  }
  return count;
}

#define COUNTENTRY "count_token"

static void storeCount(int count) {
  TDB_DATA key;
  TDB_DATA buffer;

  key.dptr = COUNTENTRY;
  key.dsize = strlen(COUNTENTRY)+1;
  buffer.dptr = (char*)&count;
  buffer.dsize = sizeof(int);
  tdb_store(dbf, key, buffer, TDB_REPLACE);
}

/**
 * Get the number of entries in the database.
 **/
int countContentEntries_TDB() {
  TDB_DATA key;
  TDB_DATA buffer;
  int count;

  key.dptr = COUNTENTRY;
  key.dsize = strlen(COUNTENTRY)+1;

  buffer = tdb_fetch(dbf, key);
  if ( (!buffer.dptr) || 
       (buffer.dsize != sizeof(int)) ) {
    count = forEachEntryInDatabase_TDB(NULL, NULL);  
    storeCount(count);
  } else {
    count = *(int*) buffer.dptr;
    free(buffer.dptr);
  }
  return count;
}

/**
 * Read the contents of a bucket to a buffer.
 *
 * @param fn the hashcode representing the entry
 * @param result the buffer to write the result to 
 *        (*result should be NULL, sufficient space is allocated)
 * @return the number of bytes read on success, -1 on failure
 **/ 
int readContent_TDB(HashCode160 * name,
		    void ** result) {
  HexName fn;
  TDB_DATA key, buffer;

  hash2hex(name, &fn);  
  key.dptr = fn.data;
  key.dsize = strlen(key.dptr) + 1;

  buffer = tdb_fetch(dbf, key);
  if (!buffer.dptr) 
    return -1;
  *result = buffer.dptr;
  return buffer.dsize;  
}

/**
 * Write content to a file. Check for reduncancy and eventually
 * append.
 *
 * @param fn the key for the entry
 * @param len the size of the block
 * @param block the data to store
 * @return SYSERR on error, OK if ok.
 **/
int writeContent_TDB(HashCode160 * name, 
		     int len,
		     void * block) {
  HexName fn;
  TDB_DATA buffer, key;

  hash2hex(name, &fn);
  key.dptr = fn.data;
  key.dsize = strlen(key.dptr) + 1;
  buffer.dptr = block;
  buffer.dsize = len;
  tdb_store(dbf, key, buffer, TDB_REPLACE);
  storeCount(countContentEntries_TDB() + 1);
  return OK;
}

/**
 * Free space in the database by removing one file
 * @param name hashcode representing the name of the file (without directory)
 **/
int unlinkFromDB_TDB(HashCode160 * name) {
  TDB_DATA key;
  HexName fn;

  hash2hex(name, &fn);
 
  key.dptr = fn.data;
  key.dsize = strlen(key.dptr) + 1;
  tdb_delete(dbf, key);
  storeCount(countContentEntries_TDB() - 1);
  return OK;
}

#endif

/* end of contentdatabase_tdb.c */
