/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * Tool to insert files into GNUnet's AFS.
 * @author Christian Grothoff
 * @author Krista Bennett
 * @file applications/afs/tools/gnunet-insert.c 
 **/

#include "config.h"
#include "applications/afs/afs.h"
#include "applications/afs/encoding/insertutil.h"
#ifdef USE_EXTRACTOR
#include "applications/afs/encoding/keywords.h"
#endif
#include "applications/afs/database/lookup.h"
#include "applications/common/port.h"
#include "applications/common/printhelp.h"
#include "util/returnvalues.h"
#include "util/logging.h"
#include "util/getopt.h"
#include "util/cron.h"
#include "util/tcpio.h"
#include "util/configuration.h"
#include "util/initialize.h"
#include <unistd.h>

/**
 * Print progess message.
 **/
static void printstatus(ProgressStats * stats,
			void * verboselevel) {
  if (*(int*)verboselevel == YES)
    printf("%8d of %8d bytes inserted\n",
	   stats->progress,
	   stats->filesize);  
}

/**
 * Prints the usage information for this command if the user errs.
 * Aborts the program.
 **/
static void printhelp() {
  printf("USAGE: gnunet-insert [OPTIONS] -f FILENAME [DESCRIPTION [KEYWORD]*]\n"
	 "-V, --verbose                 : "
	 "be verbose\n"
	 "-f FILENAME, --file=FILENAME  : "
	 "insert FILENAME (obligatory)\n"
	 "-p PRIORITY, --prio=PRIORITY  : "
	 "priority of the inserted content (default: %d)\n"
	 "-n, --noindex                 : "
	 "do not index, full insertion (full copy!)\n"
#ifdef USE_EXTRACTOR
	 "-x, --noextraction            : "
	 "disable automatic keyword extraction\n"
#endif
	 "-m MIMETYPE, --mime=MIMETYPE  : "
	 "set mimetype to be MIMETYPE\n"
	 "-r FILENAME, --root=FILENAME  : "
	 "publish FILENAME as the name of the file\n",
	 getConfigurationInt("GNUNET-INSERT",
			     "CONTENT-PRIORITY"));
  printHelpDefault();
}

static int parseOptions(int argc,
			char ** argv) {
  int c;

  freeNonNull(setConfigurationString("GNUNETD",
				     "LOGFILE",
				     NULL),
	      "gnunet-insert always use option d");
  while (1) {
    int option_index=0;
    static struct option long_options[] = {
      LONG_DEFAULT_OPTIONS,
      { "noindex",       0, 0, 'n' },
      { "file",          1, 0, 'f' },
      { "prio",          1, 0, 'p' },
      { "root",          1, 0, 'r' },
      { "mime",          1, 0, 'm' },
#ifdef USE_EXTRACTOR
      { "noextraction",  1, 0, 'x' },
#endif
      { "verbose",       0, 0, 'V' },
      { 0,0,0,0 }
    };    
    c = getopt_long(argc,
		    argv, 
#ifdef USE_EXTRACTOR
		    "vhdc:L:H:Vnxm:r:f:p:", 
#else
		    "vhdc:L:H:Vnm:r:f:p:", 
#endif
		    long_options, 
		    &option_index);    
    if (c == -1) 
      break;  // No more flags to process
    if (YES == parseDefaultOptions(c, optarg))
      continue;
    switch(c) {
    case 'm':
      freeNonNull(setConfigurationString("GNUNET-INSERT",
					 "MIMETYPE",
					 optarg),
		  "gnunet-insert option parsing m");
      break;
#ifdef USE_EXTRACTOR
    case 'x':
      freeNonNull(setConfigurationString("GNUNET-INSERT",
					 "EXTRACT-KEYWORDS",
					 "NO"),
		  "gnunet-insert option parsing x");
      break;
#endif
    case 'r':
      freeNonNull(setConfigurationString("GNUNET-INSERT",
					 "FILENAMEROOT",
					 optarg),
		  "gnunet-insert option parsing r");
      break;
    case 'V':
      freeNonNull(setConfigurationString("GNUNET-INSERT",
					 "VERBOSE",
					 "YES"),
		  "gnunet-insert option parsing V");
      break;
    case 'f': {
      char * root;
      freeNonNull(setConfigurationString("GNUNET-INSERT",
					 "FILENAME",
					 optarg),
		  "gnunet-insert option parsing f");
      root = getConfigurationString("GNUNET-INSERT",
				    "FILENAMEROOT");
      if (root == NULL) {
	/* if filename is '/home/user/foo', use 'foo' as the filenameRoot */
	int i;
	root = optarg;
	for (i=0;i<strlen(optarg);i++)
	  if (optarg[i] == '/')
	    root = &optarg[i+1];
      }
      setConfigurationString("GNUNET-INSERT",
			     "FILENAMEROOT",
			     root);
      break;
    }
    case 'p': {
      unsigned int contentPriority;
      
      if (1 != sscanf(optarg, "%ud", &contentPriority)) {
	LOG(LOG_FAILURE,
	    "FAILURE: You must pass a number to the -p option.\n");
	return SYSERR;
      }
      setConfigurationInt("GNUNET-INSERT",
			  "CONTENT-PRIORITY",
			  contentPriority);
      break;
    }
    case 'v': 
      printf("GNUnet v%s, gnunet-insert v%s\n",
	     VERSION, 
	     AFS_VERSION);
      return SYSERR;
    case 'h': 
      printhelp(); 
      return SYSERR;
    case 'n':
      freeNonNull(setConfigurationString("GNUNET-INSERT",
					 "INDEX-CONTENT",
					 "NO"),
		  "gnunet-insert parse options n");
      break;
    default: 
      LOG(LOG_FAILURE,
	  "FAILURE: Unknown option %c. Aborting.\n"\
	  "Use --help to get a list of options.\n",
	  c);
      return SYSERR;
    } /* end of parsing commandline */
  } /* while (1) */

  if (argc - optind > 0)  
    setConfigurationString("GNUNET-INSERT",
			   "DESCRIPTION",
			   argv[optind++]);
  setConfigurationStringList(&argv[optind],
			     argc - optind);
  return OK;
}


/**
 * The main function to insert files into GNUnet.
 * @param argc number of arguments from the command line
 * @param argv command line arguments
 * @return return 0 for ok, -1 on error
 **/   
int main(int argc, char ** argv) {
  int res;
  int beVerbose;
  GNUNET_TCP_SOCKET * sock;
  Block * top;
  char * filename;
  char * description;
  char * filenameRoot; /* file name for the root-node */
  char * mimetype;
  char ** keywords;
  int num_keywords;
  cron_t startTime;
  
  if (SYSERR == initUtil(argc, argv, &parseOptions)) 
    return 0;
  
  cronTime(&startTime);
  filename = getFileName("GNUNET-INSERT",
			 "FILENAME",
			 "ERROR: you must specify a filename (option -f)\n");
  mimetype = getConfigurationString("GNUNET-INSERT",
				    "MIMETYPE");
  filenameRoot = getConfigurationString("GNUNET-INSERT",
					"FILENAMEROOT");
  description = getConfigurationString("GNUNET-INSERT",
				       "DESCRIPTION");
  num_keywords = getConfigurationStringList(&keywords);
  sock = getClientSocket();
#ifdef USE_EXTRACTOR
  if (testConfigurationString("GNUNET-INSERT",
			      "EXTRACT-KEYWORDS",
			      "YES") == YES) 
    extractKeywords(filename,
		    &description,
		    &mimetype,
		    &keywords,
		    &num_keywords);
#endif
  if (num_keywords <= 0) {
#ifdef USE_EXTRACTOR
    if (testConfigurationString("GNUNET-INSERT",
				"EXTRACT-KEYWORDS",
				"YES") == YES) 
      LOG(LOG_ERROR, 
	  "ERROR: Automatic keyword extraction failed, please specify manually.\n");
    else
#endif
      LOG(LOG_ERROR, 
	  "ERROR: Not enough arguments (you must specify keywords).\n");
    printhelp();
    return -1;
  }  
  if (YES == testConfigurationString("GNUNET-INSERT",
				     "VERBOSE",
				     "YES")) {
    int k;
    beVerbose = YES;
    printf("Using keywords:\n");
    for (k=0;k<num_keywords;k++)
      printf("%3d: %s\n",
	     k,
	     keywords[k]);
  }

  if (description == NULL)
    description = "No description supplied";
  if (mimetype == NULL)
    mimetype = "unknown";
  top = insertFile(sock,
		   filename,
		   &printstatus,
		   &beVerbose);
  if (top != NULL) {
    res = insertRoot(sock,
		     top,
		     description,
		     filenameRoot,
		     mimetype,
		     num_keywords,
		     keywords);
  } else {
    LOG(LOG_DEBUG,
	"DEBUG: error in insertFile\n");
    res = SYSERR;
  }
  
  if (res == OK) {
    if(beVerbose == YES) {
      HexName hex;
      HexName hex2;
      
      hash2hex(&top->chk.key, 
	       &hex);
      hash2hex(&top->chk.query, 
	       &hex2);
      printf("File %s successfully inserted -- %s %s %X %u\n", 
	     filename,
	     (char*)&hex, 
	     (char*)&hex2,
	     crc32N(top->data, top->len), 
	     top->filesize);
      printf("Speed was %8.3f kilobyte per second.\n",
	     (top->filesize/1024.0) / 
	     (((double)(cronTime(NULL)-startTime)) / (double)cronSECONDS) );
    }    
  } else
    printf("Error inserting file %s.\n",
	   filename);
  if (top != NULL)
    top->done(top, NULL);
  releaseClientSocket(sock); 
  doneUtil();
  if (res == OK)
    return 0;
  else 
    return -1;
}  

/* end of gnunet-insert.c */
