/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * Implementation of the APIs to the GNUnet core
 * @file server/core.c
 * @author Christian Grothoff
 **/

#include "server/coreapi.h"
#include "server/core.h"
#include "util/configuration.h"
#include "util/cron.h"
#include "util/statistics.h"
#include "server/pingpong.h"
#include "server/handler.h"
#include "server/trustservice.h"
#include "server/knownhosts.h"
#include "server/keyservice.h"
#include "server/dso.h"

typedef struct ShutdownList {
  void * library;  
  char * dsoName;
  struct ShutdownList * next;
} ShutdownList;

typedef void (*ApplicationDoneMethod)();

/** globals for the major APIs **/
static CoreAPIForTransport transportCore;
static CoreAPIForApplication applicationCore;
static ShutdownList * shutdownList = NULL;

#define DSO_PREFIX "libgnunet"

/**
 * @param signer the identity of the host that presumably signed the message
 * @param message the signed message
 * @param size the size of the message
 * @param sig the signature
 * @return OK on success, SYSERR on error (verification failed)
 **/
static int verifySigHelper(const HostIdentity * signer,
			   void * message,
			   int size,
			   Signature * sig) {
  HELO_Message * helo;
  int res;

  if (SYSERR == identity2Helo(signer,
			      ANY_PROTOCOL_NUMBER,
			      &helo))
    return SYSERR;
  res = verifySig(message, size, sig,
		  &helo->publicKey);
  xfree(helo,
	"verifySigHelper: helo no longer needed");
  return res;
}

typedef struct fENHWrap {
  PerNodeCallback method;
  void * arg;
} fENHWrap;

static void fENHCallback(BufferEntry * be,
			 void * arg) {
  fENHWrap * wrap;

  wrap = (fENHWrap*) arg;
  if (wrap->method != NULL)
    wrap->method(&be->session.sender,
		 wrap->arg);
}

/**
 * Wrapper around forAllConnectedHosts...
 **/
static int forEachNodeHelper(PerNodeCallback method,
			     void * arg) {
  fENHWrap wrap;

  wrap.method = method;
  wrap.arg = arg;
  return forAllConnectedHosts(&fENHCallback,
			      &wrap);
}

/* **************** inbound message queue for all transports ************** */

/**
 * How many incoming packages do we have in the buffer
 * (max.). Must be >= THREAD_COUNT to make sense.
 **/
#define QUEUE_LENGTH 16

/**
 * How many threads do we start? 
 **/
#define THREAD_COUNT 2


static MessagePack * bufferQueue_[QUEUE_LENGTH];
static int bq_firstFree_;
static int bq_lastFree_;
static int bq_firstFull_;

static Semaphore * bufferQueueRead_;
static Semaphore * bufferQueueWrite_;   
static Mutex globalLock_;
static Semaphore * mainShutdownSignal = NULL;
static pthread_t threads_[THREAD_COUNT];

/**
 * This is the main loop of each thread.  It loops *forever* waiting
 * for incomming packets in the packet queue. Then it calls "handle"
 * (defined in handler.c) on the packet.
 **/
static void threadMain(int id) {
  MessagePack * mp;
  
  while (mainShutdownSignal == NULL) {
    semaphore_down(bufferQueueRead_);
    /* handle buffer entry */
    /* sync with other handlers to get buffer */
    if (mainShutdownSignal != NULL)
      break;
    MUTEX_LOCK(&globalLock_);
    mp = bufferQueue_[bq_firstFull_++];
    if (bq_firstFull_ == QUEUE_LENGTH)
      bq_firstFull_ = 0;
    MUTEX_UNLOCK(&globalLock_);
    /* end of sync */

    /* handle buffer - now out of sync */    
    handleMessage(mp->tsession,
		  &mp->sender,
		  mp->msg,
		  mp->size,
		  mp->isEncrypted,
		  mp->crc);
    if (mp->tsession != NULL)
      transportDisconnect(mp->tsession);
    xfree(mp->msg,
	  "freeing message buffer");
    xfree(mp,
	  "freeing message pack");
    /* free buffer */
    MUTEX_LOCK(&globalLock_);
    bufferQueue_[bq_lastFree_++] = NULL;
    if (bq_lastFree_ == QUEUE_LENGTH)
      bq_lastFree_ = 0;
    MUTEX_UNLOCK(&globalLock_);
    /* buffer is back in pool. */
    /* handling done - open buffer for write */    
    semaphore_up(bufferQueueWrite_);    
  }
  semaphore_up(mainShutdownSignal);
} /* end of threadMain */

/**
 * Processing of a message from the transport layer
 * (receive implementation).
 **/
static void receive(MessagePack * mp) {
  semaphore_down(bufferQueueWrite_);
  /* aquire buffer */
  if (SYSERR == transportAssociate(mp->tsession))
    mp->tsession = NULL;
  
  MUTEX_LOCK(&globalLock_);
  if (bq_firstFree_ == QUEUE_LENGTH)
    bq_firstFree_ = 0;
  bufferQueue_[bq_firstFree_++] = mp;    
  MUTEX_UNLOCK(&globalLock_);
  semaphore_up(bufferQueueRead_);  
}

/**
 * Initialize the CORE's globals.
 **/
void initCore() {
  int i;

  MUTEX_CREATE(&globalLock_);
  for (i=0;i<QUEUE_LENGTH;i++)
    bufferQueue_[i] = NULL;
  bq_firstFree_ = 0;
  bq_lastFree_ = 0;
  bq_firstFull_ = 0;
  
  /* initialize sync mechanisms for message handling threads */
  bufferQueueRead_ = new_semaphore(0);
  bufferQueueWrite_ = new_semaphore(QUEUE_LENGTH);
  /* create message handling threads */
  for (i=0;i<THREAD_COUNT;i++) {
    PTHREAD_CREATE(&threads_[i], 
		   (void * (*)(void *)) &threadMain,
		   (void *) &i,
		   8 * 1024); 
    pthread_detach(threads_[i]);
  }

  transportCore.version = 0;
  transportCore.myIdentity = &myIdentity; /* keyservice.c */
  transportCore.receive = &receive; /* core.c */

  applicationCore.version = 0;
  applicationCore.myIdentity = &myIdentity; /* keyservice.c */
  applicationCore.pingAction = &pingAction; /* pingpong.c */
  applicationCore.sign = &signData; /* keyservice.c */
  applicationCore.verifySig = &verifySigHelper; /* core.c */
  applicationCore.changeTrust = &changeHostCredit; /* trustservice.c */
  applicationCore.getTrust = &getHostCredit; /* trustservice.c */
  applicationCore.sendToNode = &unicast; /* connection.c */
  applicationCore.forAllConnectedNodes = &forEachNodeHelper; /* core.c */
  applicationCore.broadcastToConnected = &broadcast; /* connection.c */
  applicationCore.registerSendCallback = &registerSendCallback; /* connection.c */
  applicationCore.unregisterSendCallback = &unregisterSendCallback; /* connection.c */
  applicationCore.registerClientHandler = &registerCSHandler; /* tcpserver.c */
  applicationCore.unregisterClientHandler = &unregisterCSHandler; /* tcpserver.c */
  applicationCore.registerClientExitHandler = &registerClientExitHandler; /* tcpserver.c */
  applicationCore.unregisterClientExitHandler = &unregisterClientExitHandler; /* tcpserver.c */
  applicationCore.registerHandler = &registerp2pHandler; /* handler.c*/
  applicationCore.unregisterHandler = &unregisterp2pHandler; /* handler.c*/
  applicationCore.estimateNetworkSize = &estimateNetworkSize; /* knownhosts.c */
  applicationCore.computeIndex = &computeIndex; /* connection.c */
}

CoreAPIForTransport * getCoreAPIForTransport() {
  return &transportCore;
}

CoreAPIForApplication * getCoreAPIForApplication() { 
  return &applicationCore;
}

void loadApplicationModules() {
  CoreAPIForApplication * capi;
  ApplicationMainMethod mptr;
  char * dso;
  char * next;
  char * pos;
  ShutdownList * nxt;
  
  capi = getCoreAPIForApplication();  
  dso = getConfigurationString("GNUNETD",
			       "APPLICATIONS");
  if (dso != NULL) {
    LOG(LOG_DEBUG,
	"DEBUG: loading applications %s\n",
	dso);
    next = dso;
    do {
      pos = next;
      while ( (*next != '\0') &&
	      (*next != ' ') )
	next++;
      if (*next == '\0')
	next = NULL; /* terminate! */
      else {
	*next = '\0'; /* add 0-termination for pos */
	next++;
      }
      nxt = xmalloc(sizeof(ShutdownList),
		    "shutdown list");
      nxt->next = shutdownList;
      nxt->dsoName = strdup(pos);
      nxt->library = loadDynamicLibrary(DSO_PREFIX,
					pos);
      shutdownList = nxt;
      mptr = bindDynamicMethod(nxt->library,
			       "initialize_",
			       pos);
      mptr(capi);
    } while (next != NULL);
    xfree(dso,
	  "main::dso");
  } else
    LOG(LOG_WARNING,
	"WARNING: No application (DSO) defined in configuration!\n");
}

/**
 * Shutdown the CORE modules (also shuts down all
 * application modules).
 **/
void doneCore() {
  ShutdownList * nxt;
  ApplicationDoneMethod mptr;
  int i;

  mainShutdownSignal = new_semaphore(0);
  for (i=0;i<THREAD_COUNT;i++) {
    semaphore_up(bufferQueueRead_);
    semaphore_down(mainShutdownSignal);
  }
  free_semaphore(mainShutdownSignal);
  mainShutdownSignal = NULL;
  MUTEX_DESTROY(&globalLock_);
  free_semaphore(bufferQueueRead_);
  free_semaphore(bufferQueueWrite_);
  while (shutdownList != NULL) {
    mptr = bindDynamicMethod(shutdownList->library,
			     "done_",
			     shutdownList->dsoName);
    if (mptr != NULL)
      mptr();
    unloadDynamicLibrary(shutdownList->library);
    nxt = shutdownList->next;
    xfree(shutdownList->dsoName,
	  "strdup dsoname");
    xfree(shutdownList,
	  "shutdownlist");
    shutdownList = nxt;
  }
}

/* end of core.c */
