/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * Code to maintain the list of currently known hosts (in memory
 * structure of data/hosts) and (temporary) blacklisting information
 * and a list of HELOs that are temporary unless confirmed via PONG
 * (used to give the transport module the required information for the
 * PING).
 *
 * Todo:
 * - we may want to cache more HELOs in memory
 *
 * @author Christian Grothoff
 **/ 

#include "config.h"
#include "server/knownhosts.h"
#include "server/keyservice.h"
#include "util/cron.h"
#include "util/configuration.h"
#include "util/checksum.h"
#include "util/logging.h"

typedef struct {
  HostIdentity identity;
  /** how long is this host blacklisted? **/
  cron_t until;
  /** what would be the next increment for blacklisting? (in seconds) **/
  unsigned int delta;
  /** for which protocol is this host known? */
  unsigned short protocol;
} HostEntry;

/**
 * The current (allocated) size of knownHosts
 **/
static int max_;

/**
 * The number of actual entries in knownHosts
 **/
static int count_;

/**
 * A lock for accessing knownHosts
 **/
static Mutex lock_;

/**
 * The list of known hosts.
 **/
static HostEntry * hosts_;

/**
 * Directory where the HELOs are stored in (data/hosts)
 **/
static char * networkIdDirectory;

#define MAX_TEMP_HOSTS 32

/**
 * The list of temporarily known hosts
 **/
static HELO_Message * tempHosts[MAX_TEMP_HOSTS];

/**
 * tempHosts is a ringbuffer, this is the current
 * index into it.
 **/
static int tempHostsNextSlot;

/**
 * Get the directory in which we store the hostkeys/HELOs.
 **/
static char * getHostsDirectory() {
  return getFileName("GNUNETD",
		     "HOSTS",
		     "Configuration file must specify directory for "\
		     "network identities in section %s under %s.\n");		     
}

/**
 * Get the filename under which we would store the HELO_Message
 * for the given host and protocol.
 * @return filename of the form DIRECTORY/HOSTID.PROTOCOL
 **/
static char * getHostFileName(const HostIdentity * id,
			      const unsigned short protocol) {
  HexName fil;
  char * fn;

  hash2hex(&id->hashPubKey,
	   &fil);
  fn = xmalloc(strlen(networkIdDirectory) + 
	       sizeof(HexName) + 1 + 5 + 1, /* '.' (1) + %u (5) + '\0' (1) */
	       "getHostFileName: filename");
  sprintf(fn,
	  "%s%s.%u",
	  networkIdDirectory,
	  (char*) &fil,
	  protocol);
  return fn;
}

/**
 * Initialize this module.
 **/
void initKnownhosts() {
  int i;

  for (i=0;i<MAX_TEMP_HOSTS;i++)
    tempHosts[i] = NULL;
  tempHostsNextSlot = 0;
  max_   = 128;
  count_ = 0;
  MUTEX_CREATE_RECURSIVE(&lock_);
  networkIdDirectory = getHostsDirectory();
  mkdirp(networkIdDirectory);
  hosts_ = xmalloc(sizeof(HostEntry) * max_,
		   "initKnownhosts: hostlist");	      
  cronScanDirectoryDataHosts(NULL);
  addCronJob(&cronScanDirectoryDataHosts,
	     15 * cronMINUTES,
	     15 * cronMINUTES,
	     NULL); 
}

/**
 * Shutdown the knownhosts module.
 **/
void doneKnownhosts() {
  delCronJob(&cronScanDirectoryDataHosts,
	     15 * cronMINUTES,
	     NULL);
  MUTEX_DESTROY(&lock_);
  xfree(hosts_,
	"doneKnownhosts: hostlist");
}

/**
 * Check if 2 hosts are the same.
 * @return YES if they are equal, otherwise NO
 **/
int hostIdentityEquals(const HostIdentity * first, 
		       const HostIdentity * second) {
  if ( (first == NULL) || 
       (second == NULL) )
    return NO;
  return equalsHashCode160(&first->hashPubKey,
			   &second->hashPubKey);
}

/**
 * Add a host to the temporary list.
 **/
void addTemporaryHost(HELO_Message * tmp) {
  HexName hex;
  
  IFLOG(LOG_DEBUG,
	hash2hex(&tmp->senderIdentity.hashPubKey,
		 &hex));
  LOG(LOG_DEBUG,
      "DEBUG: adding node %s to temporary list\n",
      &hex);
  MUTEX_LOCK(&lock_);   
  if (tempHosts[tempHostsNextSlot] != NULL)
    xfree(tempHosts[tempHostsNextSlot],
	  "freeing temporary helo");
  tempHosts[tempHostsNextSlot++] = tmp;
  if (tempHostsNextSlot >= MAX_TEMP_HOSTS)
    tempHostsNextSlot = 0;
  MUTEX_UNLOCK(&lock_);   
}

/**
 * Add a host to the list.
 * @param identity the identity of the host
 * @param protocol the protocol for the host
 **/
static void addHostToKnown(HostIdentity * identity,
			   unsigned short protocol) {
  int i;

  MUTEX_LOCK(&lock_);   
  for (i=0;i<count_;i++)
    if ( (hostIdentityEquals(identity,
			     &hosts_[i].identity)) &&
	 (protocol == hosts_[i].protocol) ) {
      MUTEX_UNLOCK(&lock_);   
      return; /* already there */
    }
  if (count_ == max_) { /* grow */
    HostEntry * tmp = xmalloc(sizeof(HostEntry) * max_ * 2,
			      "addHostToKnown: grow hostlist");
    memcpy(tmp, 
	   hosts_,
	   sizeof(HostEntry) * max_);
    max_ = max_ * 2;
  }
  memcpy(&hosts_[count_].identity,
	 identity,
	 sizeof(HostIdentity));
  hosts_[count_].until = 0;
  hosts_[count_].delta = 0;
  hosts_[count_].protocol = protocol;
  count_++;
  MUTEX_UNLOCK(&lock_);   
}

/**
 * Delete a host from the list.
 **/
void delHostFromKnown(const HostIdentity * identity,
		      const unsigned short protocol) {
  char * fn;
  int i;

  MUTEX_LOCK(&lock_);   
  for (i=0;i<count_;i++)
    if ( (hostIdentityEquals(identity,
			     &hosts_[i].identity)) &&
	 (protocol == hosts_[i].protocol) ) {
      memcpy(&hosts_[i],
	     &hosts_[count_-1],
	     sizeof(HostEntry));
      count_--;   
      /* now remove the file */
      fn = getHostFileName(identity, protocol);
      unlink(fn); 
      xfree(fn,
	    "filename in delHostFromKnown");
      MUTEX_UNLOCK(&lock_);   
      return; /* deleted */
    }
  MUTEX_UNLOCK(&lock_);   
}

/**
 * Bind a host addres (helo) to a hostId.
 * @param msg the verified (!) HELO message
 **/
void bindAddress(HELO_Message * msg) {
  char * fn;
  char * buffer;
  HELO_Message * oldMsg;
  int size;
  HexName hex;

  IFLOG(LOG_INFO,
	hash2hex(&msg->senderIdentity.hashPubKey,
		 &hex));
  LOG(LOG_INFO,
      "INFO: binding address of node %s.%d\n",
      &hex, 
      ntohs(msg->protocol));
  fn = getHostFileName(&msg->senderIdentity,		       
		       ntohs(msg->protocol));
  buffer = xmalloc(MAX_BUFFER_SIZE,
		   "space for reading HELO from drive\n");
  size = readFile(fn,
		  MAX_BUFFER_SIZE,
		  buffer);
  oldMsg = (HELO_Message*) buffer;
  if (size == HELO_Message_size(oldMsg)) {
    if (ntohl(oldMsg->expirationTime) > ntohl(msg->expirationTime)) {
      xfree(fn, "bindAddress: filename");
      xfree(buffer, "io buffer");
      return; /* have more recent HELO in stock */    
    }
  }
  writeFile(fn,
	    msg, 
	    HELO_Message_size(msg),
	    "644"); 
  xfree(fn, "bindAddress: filename");
  xfree(buffer, "io buffer");
  addHostToKnown(&msg->senderIdentity,
		 ntohs(msg->protocol));
}

struct TempStorage_ {
  HexName hex;
  HELO_Message * helo;
  int result;
};

/**
 * Check if the filename matches the identity that we are searching
 * for. If yes, fill it in.
 **/
static void identity2HeloHelper(char * fn,
				char * dirName,
				struct TempStorage_ * res) {
  if (strstr(fn, (char*)&res->hex.data[0]) != NULL) {
    char * fileName;
    HELO_Message buffer;
    int size;
        
    fileName = xmalloc(strlen(networkIdDirectory) + 
		       strlen(fn) + 1,
		       "identity2HeloHelper: filename");
    sprintf(fileName,
	    "%s%s",
	    networkIdDirectory, fn);
    size = readFile(fileName, 
		    sizeof(HELO_Message), 
		    &buffer);
    if (size == sizeof(HELO_Message)) {
      HELO_Message * tmp;
      tmp = xmalloc(HELO_Message_size(&buffer),
		    "reading helo from drive (identity2helo)");
      size = readFile(fileName, 
		      HELO_Message_size(&buffer),
		      tmp);
      if (size != HELO_Message_size(&buffer)) {
	LOG(LOG_WARNING, 
	    "WARNING: Removed invalid HELO file %s (filesize %d, expected %d)\n",
	    fileName, size, HELO_Message_size(&buffer));	
	unlink(fileName);      
	xfree(tmp,
	      "helo was invalid, freeing buffer");
      } else {
	if (NO == isTransportAvailable(ntohs(buffer.protocol)) ) {
	  xfree(tmp,
		"helo was for unsupported transport protocol");
	} else if (res->result == SYSERR) {
	  res->result = OK;
	  res->helo = tmp;
	} else if (randomi(transportGetCost(ntohs(tmp->protocol))) >
		   randomi(transportGetCost(ntohs(res->helo->protocol)))) {
	  xfree(res->helo, 
		"another proto is better");
	  res->helo = tmp;
	} else
	  xfree(tmp, "existing proto won");
      }
    } else {
      LOG(LOG_WARNING,
	  "WARNING File %s not large enough: only %d < %d bytes (removed)\n",
	  fileName,
	  size, sizeof(HELO_Message));
      unlink(fileName);      
    }
    xfree(fileName,
	  "identity2HeloHelper: filename");
  }
}

/**
 * Obtain the public key and address of a known host. If no specific
 * protocol is specified (ANY_PROTOCOL_NUMBER), HELOs for cheaper
 * protocols are returned with preference (randomness!).
 *
 * @param hostId the host id
 * @param protocol the protocol that we need,
 *        ANY_PROTOCOL_NUMBER if we do not care which protocol
 * @param result where to store the result
 * @returns SYSERR on failure, OK on success
 **/
int identity2Helo(const HostIdentity *  hostId,
		  const unsigned short protocol,
		  HELO_Message ** result) {  
  struct TempStorage_ tempStorage;
  char * fn;
  HELO_Message buffer;
  int size;
  int i;

  *result = NULL;
  fn = getHostFileName(hostId, protocol);
  size = readFile(fn, 
		  sizeof(HELO_Message), 
		  &buffer);
  if (size == sizeof(HELO_Message)) {
    *result = xmalloc(HELO_Message_size(&buffer),
		      "reading helo from drive (identity2helo)");
    size = readFile(fn, 
		    HELO_Message_size(&buffer),
		    *result);
    if (size != HELO_Message_size(&buffer)) {
      LOG(LOG_WARNING, 
	  "WARNING: Removed invalid HELO file %s\n",
	  fn);
      unlink(fn);      
      xfree(fn,
	    "identity2key: filename (2)");
      xfree(*result,
	    "helo was invalid, freeing buffer");
      *result = NULL;
      return SYSERR;
    }
    if (YES == isTransportAvailable(ntohs((*result)->protocol))) {
      xfree(fn,
	    "identity2key: filename (2)");
      return OK;
    } else {
      xfree(*result,
	    "helo was invalid, freeing buffer");
      *result = NULL;
    }
  } else if (size != -1) {
    LOG(LOG_WARNING, 
	"WARNING: Removed invalid HELO file %s\n",
	fn);
    unlink(fn);      
  }
  xfree(fn,
	"identity2key: filename (1)");

  /* ok, then try temporary hosts */      
  MUTEX_LOCK(&lock_);   
  for (i=0;i<MAX_TEMP_HOSTS;i++) {
    if ( (tempHosts[i] != NULL) &&
	 hostIdentityEquals(hostId,
			    &tempHosts[i]->senderIdentity) &&
	 ( (ntohs(tempHosts[i]->protocol) == protocol) ||
	   ( (protocol == ANY_PROTOCOL_NUMBER) &&
	     (YES == isTransportAvailable(tempHosts[i]->protocol)) ) ) ) {
      *result = xmalloc(HELO_Message_size(tempHosts[i]),
			"found temporary HELO match");
      memcpy(*result,
	     tempHosts[i],
	     HELO_Message_size(tempHosts[i]));	    
      MUTEX_UNLOCK(&lock_);   
      return OK;
    }
  }
  MUTEX_UNLOCK(&lock_);   
  if (protocol != ANY_PROTOCOL_NUMBER) 
    return SYSERR; /* nothing found */
  /* ok, last chance, scan directory! */
  hash2hex(&hostId->hashPubKey,
	   &tempStorage.hex);

  tempStorage.result = SYSERR;
  tempStorage.helo = NULL;
  LOG(LOG_DEBUG,
      "DEBUG: scanning directory %s for peer identity\n",
      networkIdDirectory);
  scanDirectory(networkIdDirectory,
		(DirectoryEntryCallback)&identity2HeloHelper,
		&tempStorage);
  *result = tempStorage.helo;
  return tempStorage.result;
}

/**
 * Blacklist a host. This method is called if a host
 * failed to respond to a connection attempt.
 * @param desparation how desperate are we to connect? [0,MAXHOSTS]
 * @return OK on success SYSERR on error
 **/
int blacklistHost(HostIdentity * identity,
		  int desperation) {
  int i;
  HexName hn;

  MUTEX_LOCK(&lock_);   
  for (i=0;i<count_;i++)
    if (hostIdentityEquals(identity,
			   &hosts_[i].identity)) {
      hosts_[i].delta 
	= hosts_[i].delta * 2 + randomi((desperation+1)*cronSECONDS);
      if (hosts_[i].delta > 4 * cronHOURS) 
	hosts_[i].delta = 4 *  cronHOURS+randomi(desperation+1); 
      cronTime(&hosts_[i].until);      
      hosts_[i].until += hosts_[i].delta;
      hash2hex(&identity->hashPubKey,
	       &hn);
      LOG(LOG_DEBUG, 
	  "DEBUG: Blacklisting host %s for %d seconds.\n",
	  (char*)&hn, 
	  hosts_[i].delta / cronSECONDS);
      MUTEX_UNLOCK(&lock_);   
     return OK;
    }
  MUTEX_UNLOCK(&lock_);   
  return SYSERR;
}

/**
 * Whitelist a host. This method is called if a host
 * successfully established a connection. It typically
 * resets the exponential backoff to the smallest value.
 * @return OK on success SYSERR on error
 **/
int whitelistHost(HostIdentity * identity) {
  int i;

  MUTEX_LOCK(&lock_);   
  for (i=0;i<count_;i++) {
    if (hostIdentityEquals(identity,
			   &hosts_[i].identity)) {
      hosts_[i].delta = 0;
      hosts_[i].until = 0;
      MUTEX_UNLOCK(&lock_);   
      return OK;
    }
  }
  MUTEX_UNLOCK(&lock_);   
  return SYSERR;
}

/**
 * Call a method for each known host; not reentrant!
 * @param callback the method to call for each host
 * @param now the time to use for excluding hosts due to blacklisting, use 0 
 *        to go through all hosts.
 * @param data an argument to pass to the method
 * @return the number of hosts matching
 **/
int forEachHost(HostIterator callback,
		cron_t now,
		void * data) {
  int i;
  int count = 0;

  MUTEX_LOCK(&lock_);   
  for (i=0;i<count_;i++) {
    if (hostIdentityEquals(&hosts_[i].identity,
			   &myIdentity))
      continue;
    if ( (now == 0) || 
	 (now >= hosts_[i].until) ) {
      count++;
      if (callback != NULL) 
	callback(&hosts_[i].identity, 
		 hosts_[i].protocol,
		 data);      
    }
  }
  MUTEX_UNLOCK(&lock_);   
  return count;
}

static void cronHelper(char * filename, 
		       char * dirname,
		       void * unused) {
  HostIdentity identity;
  HexName id;
  unsigned int protoNumber;

  if (2 == sscanf(filename,
		  "%40c.%u",
		  (char*)&id,
		  &protoNumber)) {
    ((char*)&id)[40] = '\0';
    hex2hash(&id, 
	     &identity.hashPubKey);
    addHostToKnown(&identity,
		   (unsigned short) protoNumber);
  } else {
    char * fullname;

    LOG(LOG_WARNING,
	"WARNING: File in %s does not match naming convention: %s (removed)\n",
	networkIdDirectory, 
	filename);

    fullname = xmalloc(strlen(filename) + strlen(networkIdDirectory) + 1,
		       "cronHelper removes bad file");
    fullname[0] = '\0';
    strcat(fullname, networkIdDirectory);
    strcat(fullname, filename);
    unlink(fullname);
    xfree(fullname,
	  "cronHelper removes bad file");
  }
}

/**
 * Get an estimate of the network size.
 * @return the estimated number of nodes, SYSERR on error
 **/
int estimateNetworkSize() {
  return count_;
}
 
/**
 * Call this method periodically to scan data/hosts for new hosts.
 **/
void cronScanDirectoryDataHosts(void * unused) {
  int count;

  LOG(LOG_CRON,
      "CRON: enter cronScanDirectoryDataHosts\n");
  count = scanDirectory(networkIdDirectory,
			&cronHelper,
			NULL);
  if (count <= 0) {
    LOG(LOG_WARNING, 
	"WARNING: scanDirectory %s returned no known hosts!\n",
	networkIdDirectory);
  }
  LOG(LOG_CRON, 
      "CRON: exit cronScanDirectoryDataHosts\n");
}

/* end of knownhosts.c */
