/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/** 
 * @file util/statuscalls.c
 * @author Tzvetan Horozov
 * @author Christian Grothoff
 *
 * Status calls implementation for load managment.
 *
 * Todo:
 * - add capability to be notified by client-code that
 *   certain values have changed (e.g. that GNUnet did
 *   transmit 1500 bytes of data). This would allow
 *   updating the statistical values without querying 
 *   the OS and make the returned values adapt faster to
 *   changing conditions (we still need to query the OS
 *   to notice other applications using the network, but
 *   at least for GNUnet-internal changes we should be
 *   able to find out by ourselves). This should be done
 *   for network (in, out) and storage (allocated, freed).
 **/

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <dirent.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#include "util/statuscalls.h"
#include "util/logging.h"
#include "util/returnvalues.h"
#include "util/xmalloc.h"
#include "util/configuration.h"
#include "util/semaphore.h"
#include "util/cron.h"
#include "util/statistics.h"

/* where to read network interface information from
   under Linux */
#define PROC_NET_DEV "/proc/net/dev"

typedef struct {
  unsigned long long last_in;
  unsigned long long last_out;
} NetworkStats;

static char ** interfacePtrs; /* Pointer to the name of each interface, has numInterfaces entries */
static int numInterfaces = 0; /* how many interfaces do we have? */

/* configuration (max up, down, load) */
static int maxNetDownBPS;
static int maxNetUpBPS;
static int maxNetLoad; /* in percent of maxNetXXXBPS */
static int maxCPULoad; /* in percent of 1 CPU */
static int lastNetResultMax = -1; /* the max load we saw last time */
static cron_t lastnettime   =  0; /* when did we check last time? */

/* tracking */
static char ** tracked_files = NULL;
static int tracked_file_count = 0;
static NetworkStats * last_net_results; /* has numInterfaces entries */
static long long spaceLeft; /* in KB */
static long long spaceUsed; /* in KB */
static DriveFullCallback * driveFullCallbacks;
static int dfc_count;

/* quota */
static unsigned int diskQuota; /* in MB! */
static int stat_handle_kb_used;
static int stat_handle_kb_remaining;
static int stat_handle_network_load;
static int stat_handle_cpu_load;

/* for async configuration update! */
static Mutex statusMutex;

/**
 * Flag to ensure that we do not start a
 * second job while we are already
 * cleaning up...
 **/
static Semaphore * signalSemaphore = NULL;

static void spaceJob() {
  int i;

  for (i=0;i<dfc_count;i++) 
    driveFullCallbacks[i](&spaceLeft);  
  statSet(stat_handle_kb_used,
	  spaceUsed);
  if (spaceLeft < 0)
    statSet(stat_handle_kb_remaining,
	    0);
  else
    statSet(stat_handle_kb_remaining,
	    spaceLeft);

  for (i=0;i<dfc_count+1;i++)
    semaphore_up(signalSemaphore);
  MUTEX_LOCK(&statusMutex);
  free_semaphore(signalSemaphore);
  signalSemaphore = NULL;
  MUTEX_UNLOCK(&statusMutex);
}

static void cronSpaceMonitor() {
  unsigned long long use;
  int i;

  LOG(LOG_CRON,
      "CRON: cronSpaceMonitor\n");
  MUTEX_LOCK(&statusMutex);
  use = 0;
  for (i=0;i<tracked_file_count;i++)
    use += getFileSize(tracked_files[i]);
  spaceLeft /* KB */
    = (diskQuota /* MB */ * 1024) - (use /* bytes */ / 1024);

  if (spaceLeft < 1024) {  /* 1 MB buffer */
    if (signalSemaphore == NULL) {
      pthread_t job;

      signalSemaphore = new_semaphore(0);
      PTHREAD_CREATE(&job,
		     (void * (*)()) &spaceJob, 
		     NULL,
		     1024*64);
      pthread_detach(job);
    }
  }
  MUTEX_UNLOCK(&statusMutex);
  LOG(LOG_CRON,
      "CRON: cronSpaceMonitor done.\n");
}

/**
 * Register a callback for drive-full monitoring
 **/
int registerDriveFullCallback(DriveFullCallback callback) {
  DriveFullCallback * tmp;

  MUTEX_LOCK(&statusMutex);
  tmp = xmalloc((dfc_count+1)*sizeof(DriveFullCallback),
		"growing DFC array");
  if (driveFullCallbacks != NULL) {
    memcpy(tmp,
	   driveFullCallbacks,
	   dfc_count * sizeof(DriveFullCallback));
    xfree(driveFullCallbacks,
	  "growing DFC array");
  }
  driveFullCallbacks = tmp;  
  driveFullCallbacks[dfc_count++] = callback;
  MUTEX_UNLOCK(&statusMutex);
  return OK;
}

/**
 * Register a callback for drive-full monitoring
 **/
int unregisterDriveFullCallback(DriveFullCallback callback) {
  int i;

  MUTEX_LOCK(&statusMutex);
  if (signalSemaphore != NULL)
    semaphore_down(signalSemaphore);
  
  for (i=0;i<dfc_count;i++)
    if (driveFullCallbacks[i] == callback) {
      driveFullCallbacks[i] = driveFullCallbacks[--dfc_count];
      MUTEX_UNLOCK(&statusMutex);
      return OK;
    }
  MUTEX_UNLOCK(&statusMutex);
  return SYSERR;
}
 
static void cronLoadUpdate(void * unused) {
  getCPULoad();
  getNetworkLoad();
}

/**
 * The following method is called in order to initialize the status calls
 * routines. After that it is safe to call each of the status calls separately
 * @return OK on success and SYSERR on error (or calls errexit).
 **/
void initStatusCalls() {
  LOG(LOG_DEBUG,
      "DEBUG: in initStatusCalls\n");
  MUTEX_CREATE_RECURSIVE(&statusMutex);
  tracked_file_count = 0;
  last_net_results = NULL; /* has numInterfaces entries */
  spaceLeft = 0; /* in KB */
  spaceUsed = 0; /* in KB */
  dfc_count = 0;
  interfacePtrs = NULL;
  last_net_results = NULL;
  tracked_files = NULL;
  driveFullCallbacks = NULL;
  stat_handle_kb_used
    = statHandle("# kb of storage used");
  stat_handle_kb_remaining
    = statHandle("# kb of storage remaining");
  stat_handle_network_load
    = statHandle("% of allowed network load");
  stat_handle_cpu_load
    = statHandle("% of allowed cpu load");
  resetStatusCalls();
  networkUsage(); 
  cpuUsage();  
  /* every hour, check space consumption */
  addCronJob(&cronSpaceMonitor,
	     10 * cronMINUTES,
	     10 * cronMINUTES,
	     NULL);
  addCronJob(&cronLoadUpdate,
	     10 * cronSECONDS,
	     10 * cronSECONDS,
	     NULL);
}

/**
 * Shutdown the status calls module.
 **/
void doneStatusCalls() {
  delCronJob(&cronSpaceMonitor,
	     1 * cronHOURS,
	     NULL);
  delCronJob(&cronLoadUpdate,
	     10 * cronSECONDS,
	     NULL);
  xfree(interfacePtrs[0], 
	"doneStatusCalls: interface-string(s)");
  if (last_net_results != NULL)
    xfree(last_net_results,
	  "doneStatusCalls: network statistics");    
  freeNonNull(tracked_files,
	      "doneStatusCalls: freeing list of tracked files");
  if (driveFullCallbacks != NULL)
    xfree(driveFullCallbacks,
	  "doneStatusCalls: freeing list of out of space callbacks");
  MUTEX_DESTROY(&statusMutex);
}

/**
 * Re-read the configuration for statuscalls.
 **/
void resetStatusCalls() {
  char * interfaces;
  char * ifcs;
  int start;

  MUTEX_LOCK(&statusMutex);
  interfaces 
    = getConfigurationString("LOAD",
			     "INTERFACES");
  ifcs = interfaces;
  /* fail if config-file is incomplete */
  if (interfaces == NULL) 
    errexit("No network interfaces defined!\n");

  /* The string containing the interfaces is formatted in the following way: 
   * each comma is replaced by '\0' and the pointers to the beginning of every 
   * interface are stored
   */
  numInterfaces = 0;
  start = YES;
  while (1) {
    if (*interfaces == '\0') {
      if (start == NO) 
	numInterfaces++;      
      break;
    }
    if ( ((*interfaces>='a') && (*interfaces<='z')) || 
	 ((*interfaces>='A') && (*interfaces<='Z')) || 
	 ((*interfaces>='0') && (*interfaces<='9')) ) {
      start = NO;         
    } else {
      if (*interfaces != ',')
	errexit("interfaces string (%s) invalid\n",
		ifcs);
      if (start == NO) {
	start = YES;
	numInterfaces++;
      }
    }
    interfaces++;
  }
  if (numInterfaces <= 0) 
    errexit("ERROR: No network interfaces specified in he configuration file\n");

  interfacePtrs = xmalloc(sizeof(char*) * numInterfaces,
			  "allocating space for interfaces");
  last_net_results = xmalloc(sizeof(NetworkStats) * numInterfaces,
			     "allocating space for interface statistics");
  memset(last_net_results, 0,
	 sizeof(NetworkStats) * numInterfaces);

  /* 2nd pass, this time remember the positions */
  interfaces = ifcs;
  numInterfaces = 0;
  start = YES;
  while (1) {
    if (*interfaces=='\0') {
      if (start == NO) 
	numInterfaces++;      
      break;
    }
    if ( ((*interfaces>='a') && (*interfaces<='z')) || 
	 ((*interfaces>='A') && (*interfaces<='Z')) || 
	 ((*interfaces>='0') && (*interfaces<='9')) ) {
      if (start == YES) {
	start = NO;   
	interfacePtrs[numInterfaces] = interfaces;	
      }
    } else {
      if (start == NO) {
	start = YES;
	*interfaces = '\0';	
	numInterfaces++;
      }
    }
    interfaces++;
  }

  maxNetDownBPS 
    = getConfigurationInt("LOAD",
			  "MAXNETDOWNBPSTOTAL");
  maxNetUpBPS 
    = getConfigurationInt("LOAD",
			  "MAXNETUPBPSTOTAL");
  maxNetLoad
    = getConfigurationInt("LOAD",
			  "MAXNETLOAD");
  maxCPULoad
    = getConfigurationInt("LOAD",
			  "MAXCPULOAD");
  diskQuota
    = getConfigurationInt("LOAD",
			  "DISKQUOTA");
  MUTEX_UNLOCK(&statusMutex);  
  cronSpaceMonitor();
}


/**
 * The following routine returns the percentage of available used
 * bandwidth.  Example: If 81 is returned this means that 81% of the
 * network bandwidth of the host is consumed.  The method
 * initStatusCalls() should be called before this routine is invoked.
 * If there is an error the method returns -1.
 **/
int networkUsage() {
#define MAX_PROC_LINE 5000
  int lastNetResultUp    = -1;
  int lastNetResultDown  = -1;
  char line[MAX_PROC_LINE];
  unsigned long long rxnew, txnew;
  unsigned long long rxdiff, txdiff;
  int i=0;
  int ifnum;
  cron_t now, elapsedtime;
  
#ifdef LINUX
  FILE * proc_net_dev;
  char * data;
#else
  FILE * command;
#endif

  MUTEX_LOCK(&statusMutex);
  /* If we checked less than 2 seconds ago, don't do
     it again, recycle the result */
  cronTime(&now);  
  elapsedtime = now - lastnettime;
  if (elapsedtime < 2 * cronSECONDS) {
    MUTEX_UNLOCK(&statusMutex);
    return lastNetResultMax;
  }
  lastnettime = now;

  /* ok, full program... */
  if ( (maxNetUpBPS == 0) || 
       (maxNetDownBPS == 0) ) {
    lastNetResultMax = -1;
    MUTEX_UNLOCK(&statusMutex);
    return -1;
  }
  rxdiff = 0;
  txdiff = 0;
  
#ifdef LINUX
  proc_net_dev = fopen(PROC_NET_DEV, "r");
  /* Try to open the file*/
  if (NULL == proc_net_dev) {
    LOG(LOG_ERROR,
	"ERROR: Could not open %s (%s)\n", 
	PROC_NET_DEV,
	strerror(errno));
    MUTEX_UNLOCK(&statusMutex);
    lastNetResultMax = -1;
    return -1;
  }
  ifnum = 0;
  /* Parse the line matching the interface ('eth0') */
  while ( (!feof(proc_net_dev)) && 
	  ( ifnum < numInterfaces) ) {    
    fgets(line,
	  MAX_PROC_LINE, 
	  proc_net_dev);
    
    for (i=0;i<numInterfaces;i++) {
      if (NULL != strstr(line, interfacePtrs[i]) ) {
	data = (char*)strchr(line, ':');
	data++;	
	if (sscanf(data,
		   "%llu %*s %*s %*s %*s %*s %*s %*s %llu",
		  &rxnew, &txnew) != 2) {
	  fclose(proc_net_dev);
	  errexit("ERROR: reading interface data from %s\n",
		  PROC_NET_DEV);
	} 	
	rxdiff += rxnew - last_net_results[ifnum].last_in;	  
	txdiff += txnew - last_net_results[ifnum].last_out;
	/*LOG(LOG_NOTHING,
	    "NOTHING: (%d<%d) %s -- rxdiff %llu = %llu - %llu; txdiff %llu = %llu - %llu\n",
	    ifnum, numInterfaces,
	    line,
	    rxdiff, rxnew, last_net_results[ifnum].last_in,
	    txdiff, txnew, last_net_results[ifnum].last_out);*/
	last_net_results[ifnum].last_in = rxnew;
	last_net_results[ifnum].last_out = txnew;
	ifnum++;
	break;
      }
    }
  }
  fclose(proc_net_dev);
#else
  if (1) {
    MUTEX_UNLOCK(&statusMutex);
    return 0;
  }
  if ( ( command = popen("netstat -n -f inet -i", "r") ) == NULL ) {
    LOG(LOG_ERROR,
	"ERROR: could not open netstat\n");
    lastNetResultMax = -1;
    MUTEX_UNLOCK(&statusMutex);
    return -1;
  }
  ifnum = 0;
  while ( (!feof(command)) && 
	  (ifnum < numInterfaces ) ) {
    fgets(line, 
	  MAX_PROC_LINE, 
	  command);
    for (i=0; i < numInterfaces; i++) {
      if ( NULL != strstr(line, interfacePtrs[i]) ) {
	if(sscanf(line, "%*s %*s %*s %*s %llu %*s %llu %*s %*s", 
		  &rxnew, &txnew) != 2 ) {
	  pclose(command);
	  errexit("ERROR: reading interface data using netstat\n");
	}
	rxdiff += rxnew - last_net_results[ifnum].last_in;
	last_net_results[ifnum].last_in = rxnew;
	txdiff += txnew - last_net_results[ifnum].last_out;
	last_net_results[ifnum].last_out = txnew;
	ifnum++;
	break;
      } /* if match */
    }  /* for all interfaces, find match */
  } /* while: for all lines in proc */
  pclose(command);
#endif
  
  lastNetResultUp 
    = (100 * txdiff * cronSECONDS) / (elapsedtime * maxNetUpBPS);
  lastNetResultDown 
    = (100 * rxdiff * cronSECONDS) / (elapsedtime * maxNetDownBPS);
	
  if (lastNetResultUp > lastNetResultDown) 
    lastNetResultMax = lastNetResultUp;
  else 
    lastNetResultMax = lastNetResultDown;  

  MUTEX_UNLOCK(&statusMutex);
  return(lastNetResultMax);  
}

/**
 * The following routine returns a number between 0-100 (can be larger than 100
 * if the load is > 1) which indicates the percentage CPU usage.  
 * 
 * Before its first invocation the method initStatusCalls() must be called.
 * If there is an error the method returns -1
 **/
int cpuUsage(){
  cron_t now, elapsedtime;
  static cron_t lastcputime  =  0; /* In seconds */
  static int lastcpuresult = -1;
#ifndef HAVE_GETLOADAVG
#ifdef LINUX
  int ret = -1;
  static int last_cpu_results[4] = { 0, 0, 0, 0 };
  FILE* proc_stat;
  char line[128];
  int user_read, system_read, nice_read, idle_read;
  int user, system, nice, idle;
  int usage_time=0, total_time=1;
#endif
#else
  double loadavg;
#endif

  MUTEX_LOCK(&statusMutex);
  cronTime(&now);  
  elapsedtime = now - lastcputime;
  if (elapsedtime < 2 * cronSECONDS) {
    MUTEX_UNLOCK(&statusMutex);
    return lastcpuresult;
  }
  lastcputime = now;
  
#ifndef HAVE_GETLOADAVG
#ifdef LINUX
  /* Get the current time in microseconds */

  proc_stat = fopen("/proc/stat", "r");
  /* Try to open the file*/
  if (NULL == proc_stat) {
    LOG(LOG_ERROR,
	"ERROR: Could not open /proc/stat");
    lastcpuresult = -1;
    MUTEX_UNLOCK(&statusMutex);
    return -1;
  }
  /* Get the first line with the data */
  if(fgets(line, 128, proc_stat)==NULL) {
    LOG(LOG_ERROR,
	"ERROR: reading /proc/stat file");
    fclose(proc_stat);
    lastcpuresult = -1;
    MUTEX_UNLOCK(&statusMutex);
    return -1;
  }
  if (sscanf(line, "%*s %i %i %i %i", 
	     &user_read, &system_read, &nice_read,
	     &idle_read) != 4) {
    fclose(proc_stat);
    LOG(LOG_ERROR,
	"ERROR: decoding /proc/stat file");
    lastcpuresult = -1;
    MUTEX_UNLOCK(&statusMutex);
    return -1;
  }

  /* Store the current usage*/
  user = user_read - last_cpu_results[0];
  system = system_read - last_cpu_results[1];
  nice = nice_read - last_cpu_results[2];
  idle = idle_read - last_cpu_results[3];
  
  /* Calculate the % usage */
  if ((user + system + nice + idle) > 0) {
    usage_time = user + system + nice;
    total_time = usage_time + idle;
  }
  ret = (100 * usage_time) / total_time;

  /* Store the values for the next calculation*/
  last_cpu_results[0] = user_read;
  last_cpu_results[1] = system_read;
  last_cpu_results[2] = nice_read;
  last_cpu_results[3] = idle_read;
  fclose(proc_stat);

  lastcpuresult = ret;
  MUTEX_UNLOCK(&statusMutex);
  return ret;
#else
  /* loadaverage not defined and not linux: error */
  lastcpuresult = -1;
  MUTEX_UNLOCK(&statusMutex);
  return -1;
#endif
#else
  if ( getloadavg(&loadavg, 1) == -1 ) {
    LOG(LOG_ERROR,
	"ERROR: getting load average\n");
    lastcpuresult = -1;
    MUTEX_UNLOCK(&statusMutex);
    return -1;
  }
  lastcpuresult = (int) (100 * loadavg);
  MUTEX_UNLOCK(&statusMutex);
  return lastcpuresult;
#endif
}

/**
 * Get the load of the network relative to what is allowed.
 * @return the network load as a percentage of allowed
 *        (100 is equivalent to full load)
 **/
int getNetworkLoad() {
  static int lastRet = 0;
  int ret = (100 * networkUsage()) / maxNetLoad;
  if (ret > lastRet) 
    lastRet = ret; /* quickly rise to higher loads */
  else {
    ret = (ret + 7 * lastRet)/8;
    lastRet = ret; /* but only slowly go back! */
  }
  statSet(stat_handle_network_load, ret);
  return ret;
}

/**
 * Get the load of the CPU relative to what is allowed.
 * @return the CPU load as a percentage of allowed
 *        (100 is equivalent to full load)
 **/
int getCPULoad() {
  static int lastRet = 0;
  int ret = (100 * cpuUsage()) / maxCPULoad;
  if (ret > lastRet) 
    lastRet = ret; /* quickly rise to higher loads */
  else {
    ret = (ret + 7 * lastRet)/8;
    lastRet = ret; /* but only slowly go back! */
  }
  statSet(stat_handle_cpu_load, ret);
  return ret;
}

/**
 * @return the number of kilobytes GNUnet is allowed to use.
 **/
unsigned long long getDiskQuota() {
  return diskQuota * 1024;
}

/**
 * @return the number of kilobytes we have left.
 **/
long long getSpaceLeft() {
  return spaceLeft;
}

/**
 * Register a file for monitoring space consumption.
 **/
int registerSpaceMonitor(char * filename) {
  char ** tmp;
  unsigned long long size;

  LOG(LOG_DEBUG,
      "DEBUG: space monitor for %s registered\n",
      filename);
  MUTEX_LOCK(&statusMutex);
  tmp = xmalloc((tracked_file_count+1) * sizeof(char*),
		"registerSpaceMonitor: growing file tracking list");
  if (tracked_file_count > 0) {
    memcpy(tmp,
	   tracked_files,
	   sizeof(char*) * tracked_file_count);
    xfree(tracked_files,
	  "registerSpaceMonitor: growing");
  }
  tracked_files = tmp;
  tracked_files[tracked_file_count++] = strdup(filename);

  size = getFileSize(filename) / 1024;
  spaceLeft -= size;
  spaceUsed += size;
  statSet(stat_handle_kb_used,
	  spaceUsed);
  if (spaceLeft < 0)
    statSet(stat_handle_kb_remaining,
	    0); /* negative numbers not allowed in stats! */
  else
    statSet(stat_handle_kb_remaining,
	    spaceLeft);
  MUTEX_UNLOCK(&statusMutex);
  return OK;
}

/* end of statusCalls.c */
